#include "ezdmahelper.h"

// ~2912 bytes per CircularBufferHelper
EzDmaHelper<512, 4, sizeof(void*)> helper;

// volatile
// AND only incremented in corresponding callback task
// AND only decremented in loop() task
static volatile uint32_t g_BuffersAvailableToRead = 0;
static volatile uint32_t g_BuffersAvailableToWrite = 0;

void onBufferWrittenCallback(void * pvParameters)
{
  static uint8_t readAvailableCount = 0;
  Serial.println("+ (buffer written)");
  Serial.flush();
  
  readAvailableCount++;
  __sync_fetch_and_add(&g_BuffersAvailableToRead, 1);
}
void onBufferReadCallback(void * pvParameters)
{
  static uint8_t writeAvailableCount = 0;
  Serial.println("+ (buffer read)");
  Serial.flush();
  
  writeAvailableCount++;
  __sync_fetch_and_add(&g_BuffersAvailableToWrite, 1);
}

void setup(){
    Serial.begin(115200);
    while (!Serial) {
      ; // wait for serial port to connect. Needed for native USB port only
    }
    // prints title with ending line break
    Serial.println("EasyDMA Helper -- test sketch");
    Serial.flush();
    delay(1000);

    helper.begin();
    helper.onBufferRead   (onBufferReadCallback,     nullptr);
    helper.onBufferWritten(onBufferWrittenCallback,  nullptr);
    g_BuffersAvailableToRead  = helper.GetBufferCountReadyForRead();
    g_BuffersAvailableToWrite = helper.GetBufferCountReadyForWrite();
    Serial.println("Setup complete");
    Serial.flush();
    delay(1000);
}


void loop() {
    static uint32_t loopCount = 0;
    loopCount++;
    if (1 == loopCount % 16) {
        Serial.print(".");
        Serial.flush();
    }
  
    // guaranteed to only get larger from callbacks...
    uint32_t buffersReadCount = 0;
    while (0 != g_BuffersAvailableToRead) {
        // interlocked/atomic decrement
        __sync_fetch_and_sub(&g_BuffersAvailableToRead, 1);
        
        Serial.println("_ Getting buffer for READ");
        Serial.flush();
        
        void* buffer = helper.GetNextBufferForRead();

        // assert it's non-null?
        if (nullptr == buffer) {
          while(1) {
            Serial.println("_ nullptr from GetNextBufferForRead()?");
            Serial.flush();
            delay(5000);
          }
        }
                
        buffersReadCount ++;
    }

    while (buffersReadCount) {
      Serial.println("_ Indicating that READ completed");
      Serial.flush();

      buffersReadCount--;
      delay(200);
      helper.SetReadCompleted();
    }

    // guaranteed to only get larger from callbacks...
    uint32_t buffersWrittenCount = 0;
    while (0 != g_BuffersAvailableToWrite) {
        // interlocked/atomic decrement
        __sync_fetch_and_sub(&g_BuffersAvailableToWrite, 1);
        Serial.println("_ Getting buffer for WRITE");
        Serial.flush();

        void* buffer = helper.GetNextBufferForWrite();

        // assert it's non-null?
        if (nullptr == buffer) {
          while(1) {
            Serial.println("_ nullptr from GetNextBufferForWrite()?");
            Serial.flush();
            delay(5000);
          }
        }
                
        buffersWrittenCount ++;
    }

    while (buffersWrittenCount) {
      Serial.println("_ Indicating that WRITE completed");
      Serial.flush();
      buffersWrittenCount--;
      delay(200);
      helper.SetWriteCompleted();
    }   
    

}
